/* Copyright (C) 1989 Aladdin Enterprises.  All rights reserved.
   Distributed by Free Software Foundation, Inc.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* gxcache.c */
/* Character cache routines for GhostScript library */
#include "gx.h"
#include "memory_.h"
#include "gxfixed.h"
#include "gxmatrix.h"
#include "gzdevice.h"
#include "gzcolor.h"
#include "gxfont.h"
#include "gxchar.h"
#include "gzstate.h"

/* Define the size of a cached_char */
int cached_char_sizeof = sizeof(cached_char);

/* Allocate storage for caching a rendered character, */
/* and set up the memory device. */
/* Return the cached_char if OK, 0 if too big. */
cached_char *
gx_alloc_char_bits(gs_font_dir *dir, gx_device_memory *dev,
  ushort iwidth, ushort iheight)
{	unsigned iraster = (iwidth + 7) >> 3;
	unsigned isize = iraster * iheight;
	cached_char *cc;
	if ( iraster != 0 && iheight > dir->upper / iraster )
		return 0;
	/* Character will fit.  Make sure there is room. */
	if ( dir->csize >= dir->cmax || dir->bmax - dir->bsize < isize )
	   {	/* There isn't enough room.  Clear the entire cache. */
		/* We'll do something better someday.... */
		if ( dir->cmax == 0 || dir->bmax < isize ) return 0;
		dir->bsize = 0;
		dir->csize = 0;
		dir->bnext = 0;
		dir->cnext = 0;
		memset((char *)dir->chars, 0, gx_char_cache_modulus * sizeof(cached_char *));
	   }
	/* Allocate the cache entry and the bits. */
	cc = &dir->cdata[dir->cnext++];
	cc->bits = &dir->bdata[dir->bnext];
	dir->bnext += isize;
	memset((char *)cc->bits, 0, isize);
	cc->width = iwidth;
	cc->raster = iraster;
	cc->height = iheight;
	dev->width = iwidth;
	dev->height = iheight;
	dev->raster = iraster;
	dev->base = cc->bits;
	dir->csize++;
	dir->bsize += isize;
	return cc;
}

/* Remove the just-allocated character from the cache. */
/* The character hasn't been added yet. */
void
gx_unalloc_cached_char(gs_font_dir *dir, cached_char *cc)
{	uint isize = cc->raster * cc->height;
	dir->cnext--;
	dir->bnext -= isize;
	dir->csize--;
	dir->bsize -= isize;
}

/* Add a character to the cache */
void
gx_add_cached_char(gs_font_dir *dir, cached_char *cc)
{	/* Add the new character at the tail of its chain. */
	cached_char **head = &dir->chars[cc->code & (gx_char_cache_modulus - 1)];
	while ( *head != 0 ) head = &(*head)->next;
	*head = cc;
	cc->next = 0;
}

/* Look up a character in the cache. */
/* Return the cached_char or 0. */
cached_char *
gx_lookup_cached_char(gs_font_dir *dir,
  int code, gs_font *font, gs_matrix *pmat)
{	float mxx = pmat->xx, myy = pmat->yy;
	register cached_char *cc =
		dir->chars[code & (gx_char_cache_modulus - 1)];
	while ( cc != 0 )
	   {	if (	cc->code == code && cc->font == font &&
			cc->mxx == mxx && cc->myy == myy
		   )
		  return cc;
		cc = cc->next;
	   }
	return 0;
}

/* Copy a cached character to the screen. */
/* Assume the caller has already done gx_color_render (or gx_color_load). */
/* If do_it is false, just check whether the copy would succeed. */
/* Return 0 if OK, 1 if wouldn't succeed, or a negative error code. */
int
gx_copy_cached_char(gs_state *pgs, cached_char *cc, int do_it)
{	gs_fixed_point cpt;
	gs_color *pcolor = pgs->color;
	fixed x, y;
	int w, h;
	int code;
	/* If a halftone is involved, we can't do it. */
	if ( !color_is_pure(pcolor) ) return 1;
	if ( (code = gx_path_current_point(pgs->path, &cpt)) < 0 )
	  return code;
	/* If the character doesn't lie entirely within the */
	/* quick-check clipping rectangle, we can't do it. */
	x = fixed_rounded(cpt.x - cc->offset.x);
	y = fixed_rounded(cpt.y - cc->offset.y);
	w = cc->width;
	h = cc->height;
	if ( !gx_cpath_includes_rectangle(pgs->clip_path,
			x, y,
			x + int2fixed(w), y + int2fixed(h)) )
		return 1;
	if ( !do_it )
		return 0;
	/* Really do the copy */
	   {	gx_device *dev = pgs->device->info;
		code = (*dev->procs->copy_mono)
			(dev, cc->bits, 0, cc->raster,
			 fixed2int(x), fixed2int(y), w, h,
			 -1, pcolor->dev_color.color1);
	   }
	return ( code < 0 ? code : 0 );
}
